class InputValidator:
    """
    A class that validates the input parameters for a scanning process.

    Attributes:
        type (str): The type of the scanning process. Can be "continuous" or "discrete".
        scan_edges (tuple): The starting and ending positions of the scan in mm.
        stepsize (float): The size of the step between two consecutive scan positions in mm.
        velocity (float): The velocity of the scanner in mm/s.
        acceleration (float): The acceleration of the scanner in mm/s^2.

    Methods:
        validate_type: Validates the type of the scanning process.
        validate_scan_edges: Validates the scan edges.
        validate_stepsize: Validates the step size of the scanning process.
        validate_velocity: Validates the velocity of the scanner.
        validate_acceleration: Validates the acceleration of the scanner.
        validate: Validates all the input parameters.
    """
    def __init__(self, scan_pars):
        """
        Initializes the InputValidator class.

        Args:
            scan_pars (dict): A dictionary containing the input parameters for the scanning process.
                It must contain the following keys:
                    "type" (str): type of the scan, either "continuous" or "discrete".
                    "scan_edges" (list): a list containing the starting and final points of the scan.
                    "stepsize" (float): the step size between scan points.
                    "velocity" (float): the velocity of the scan motion in mm/s.
                    "acceleration" (float): the acceleration of the scan motion in mm/s^2.
        """
        self.type = scan_pars["type"]
        self.scan_edges = scan_pars["scan_edges"]
        self.stepsize = scan_pars["stepsize"]
        self.velocity = scan_pars["velocity"]
        self.acceleration = scan_pars["acceleration"]
        
    def validate_type(self):
        """
        Validates the type of the scanning process.

        Raises:
            ValueError: If the type is not "continuous" or "discrete".
        """
        if not (self.type == "continuous" or self.type == "discrete"):
            print("Invalid input: type must be either continuous or discrete.")
            raise ValueError
            
    def validate_scan_edges(self):
        """
        Validates the scan edges.

        Raises:
            ValueError: If the scan edges are out of range [0, 102].
        """
        if ((self.scan_edges[0] < 0.) | (self.scan_edges[0] > 102.)):
            print("Invalid input: first scan edge if out of range [0,102].")
            raise ValueError
        if ((self.scan_edges[1] < 0.) | (self.scan_edges[1] > 102.)):
            print("Invalid input: first scan edge if out of range [0,102].")
            raise ValueError
            
    def validate_stepsize(self):
        """
        Validates the step size of the scanning process.

        Raises:
            ValueError: If the step size is out of range [0.0005, 102], or the starting point plus step size are
                        out of range.
        """
        if ((self.stepsize < 0.5e-3) | (self.stepsize >= 102)):
            print("Invalid input: stepsize is out of range [0.0005,102] .")
            raise ValueError
        if (self.scan_edges[0]<self.scan_edges[1]):
            if (self.stepsize + self.scan_edges[0] > 102):
                print("Invalid input: first scan edge plus stepsize is of range.")
                raise ValueError
        elif (self.scan_edges[0] > self.scan_edges[1]):
            if (self.scan_edges[0] - self.stepsize < 0):
                print("Invalid input: first scan edge plus stepsize is of range.")
                raise ValueError
        else:
            print("Invalid input: the scan edges coincide.")
            raise ValueError
            
    def validate_velocity(self):
        """
        Validates the velocity attribute to ensure that it is within the range [0,10] mm/s.

        Raises:
            ValueError: If the velocity value is out of range [0,10] mm/s.
        """
        if ((self.velocity < 0) | (self.velocity > 10)):
            print("Invalid input: velocity value is out of range [0,10] mm/s.")
            raise ValueError
            
    def validate_acceleration(self):
        """
        Validates the acceleration attribute to ensure that it is within the range [0,20] mm/s^2.

        Raises:
            ValueError: If the acceleration value is out of range [0,20] mm/s^2.
        """
        if ((self.acceleration < 0) | (self.acceleration > 20)):
            print("Invalid input: acceleration value is out of range [0,20] mm/s^2.")
            raise ValueError
    
    def validate(self):
        """
        Calls the individual validation methods to validate all the scan parameters.
        If all the parameters pass the validation, it returns True.
        
        Returns:
            bool: True, if all the scan parameters pass the validation.
        Raises:
            ValueError: If any of the scan parameters fail the validation.
        """
        self.validate_type()
        self.validate_scan_edges()
        self.validate_stepsize()
        self.validate_velocity()
        self.validate_acceleration()
        return True



import numpy as np

class InputProcessor():
    """A class to process input scan parameters and provide the correct parameters to set
    in the Zurich Lock-In API data acquisition."""
    
    def __init__(self,scan_pars):
        """
        Initialize an InputProcessor instance.

        Parameters:
        -----------
        scan_pars : dict
            A dictionary containing scan parameters, including:
                - scan_edges : list of two floats
                    The edges of the scan range.
                - stepsize : float
                    The step size for the scan.
                - acceleration : float
                    The acceleration of moving stage scans.
                - velocity : float
                    The velocity of the moving stage scans.
                - sampling_frequency : float
                    The sampling frequency used for discrete scans.
                - type : str
                    The type of scan, either 'continuous' or 'discrete'.
        delta : float
            a float containing the absolute value of the distance between the scan edges.
        """
        self.scan_edges = scan_pars["scan_edges"]
        self.stepsize = scan_pars["stepsize"]
        self.acc = scan_pars["acceleration"]
        self.vel = scan_pars["velocity"]
        self.sampl_freq = scan_pars["sampling_freq"]
        self.type = scan_pars["type"]
        self.delta = self.delta_calculator()

    
    def rows_columns_contiuous(self):
            """
            Process input data to find the number of rows and columns
            for a continuous scan with PI controller and Zurich lock-in.

            Returns:
            --------
            N_cols : int
                The number of columns for the data acquisition.
            N_rows : int
                The number of rows for the data acquisition.
            """
            N_rows = 1
            N_cols = int(np.floor(self.delta/self.stepsize)) + 1
            return N_rows, N_cols
    
    def rows_columns_discrete(self):
        """
        Process input data to find the number of rows and columns
        for a discrete scan with Pi controller and Zurich lock-in

        Returns:
        --------
        N_cols : int
            The number of columns for the data acquisition.
        N_rows : int
            The number of rows for the data acquisition.
        """
        N_rows = int(np.floor(self.delta/self.stepsize)) + 1
        N_cols = int(np.floor(0.05*self.sampl_freq))
        return N_rows, N_cols

    def evaluate_daq_pars(self):
        """
        Process input data for a 1D scan to return the complete DAQ parameters
        that suit the scan parameters.

        Returns:
        --------
        daq_pars : dict
            A dictionary containing data acquisition parameters, including:
                - daq_columns : int
                    The number of columns for the data acquisition.
                - daq_rows : int
                    The number of rows for the data acquisition.
                - duration : float
                    The duration of the triggered data acquisition.
                - mode : str
                    The data acquisition mode; 'Exact (on-grid)' for discrete scan, 'Linear' for continuous
                - trigger type : str
                    The trigger type, always set to 'HW trigger'.
                - trigger edge : str
                    The trigger edge; 
                - holdoff : float
                    The holdoff time for the trigger
        """

        if self.type == "continuous":
            duration = self.duration_calculator()
            N_rows, N_cols = self.rows_columns_contiuous()
            mode = "linear"
            edge = "positive"
        else:
            duration = 0.05 # 50 milliseconds
            N_rows, N_cols = self.rows_columns_discrete()
            mode = "Exact (on-grid)"
            edge = "negative"
            
        daq_pars =  {
                    "daq_columns" : N_cols,
                    "daq_rows" : N_rows,
                    "duration" : duration,
                    "mode" : mode,
                    "trigger type" : "HW trigger",
                    "trigger edge" : edge,
                    "holdoff" : duration*(0.95),
                    }
        return daq_pars

    def delta_calculator(self):
        """
        Calculate the absolute value of the distance between the scan edges.

        Returns:
        --------
        delta : float
            Absolute value of the distance between scan edges.
        """
        return abs(self.scan_edges[1]-self.scan_edges[0])

    def duration_calculator(self):
        """
        Calculate the duration of the triggered data acquisition for a continuous scan.

        Parameters:
        -----------
        acc : float
            The acceleration of the moving stage scans.
        delta : float
            The distance between the scan edges.
        vel : float
            The velocity of the moving stage scans.

        Returns:
        --------
        duration : float
            The duration of the triggered data acquisition.
        """
        if (np.sqrt(self.acc*self.delta)>self.vel):
            duration = self.vel/self.acc + self.delta/self.vel
        else:
            duration = 2*np.sqrt(self.delta/self.acc)
        return duration

            
        

    


from pizurscan.PI_commands import Stepper
from pizurscan.Scanner import Scanner
from pizurscan.InputProcessor import InputProcessor
from pipython import pitools,GCS2Commands,GCSDevice
from pizurscan.OutputProcessor import OutputProcessor
from pizurscan.InputValidator import InputValidator
import pytest as pytest
from math import isclose
import numpy as np 

class TestStepper:
    stepper = Stepper('C-663', 'L-406.40SD00')
    
    def test_initialization_pidevice(self):
        """Tests that when a Stepper object is initialiazed through 
        the __init__ method, the instantiation of the GCSDevice is corrected."""
        assert isinstance(self.stepper.pidevice, GCSDevice)
    
    def test_initialization_controller_axis(self):
        """Tests that when a Stepper object is initialized through
        the __init__ method, the attributes controller_id and axis_id are
        properly assigned."""
        controller = 'C-663'
        axis = 'L-406.40SD00'
        # make id of controller and axis attributes
        assert self.stepper.controller_id == controller
        assert self.stepper.axis_id == axis
            
    def test_usb_return_list(self):
        """
        Tests that when a Stepper object is passed to 
        usb_plugged_devices the returned object is a list.
        """
        assert isinstance(self.stepper.usb_plugged_devices(), list)
    
    def test_usb_return_notempty_list(self):
        """Test that when a Stepper object is passed to 
        usb_plugged_devices, the returned list is not empty."""
        assert self.stepper.usb_plugged_devices() != []
        
    def test_connection_is_established(self):
        """
        Tests that when a Stepper object is passed to connect_pidevice,
        a connection to the controller is successfully established.
        """
        self.stepper.connect_pidevice()
        assert self.stepper.pidevice.IsConnected()
  
    def test_connected_axis(self):
        """
        Tests that when a Stepper object is connected, the connected axis 
        is one ('1'). C663 controller supports connection with only 
        axis at a time. 
        """
        assert pitools.getaxeslist(self.stepper.pidevice, axes=None) == ['1']
        
    def test_pos_is_float(self):
        """Tests that when the current position of a connected Stepper object is probed, 
        the position returned by get_curr_pos is a float number."""
        pos = self.stepper.get_curr_pos()['1']
        assert isinstance(pos, float)
    
    def test_pos_is_in_range(self):
        """Tests that when the current position of a connected Stepper object is probed, 
        the position returned by get_curr_pos is comprised in the available range (0,102)"""
        pos = self.stepper.get_curr_pos()['1']
        assert pos >= 0.
        assert pos <= 102.
        
    def test_negative_reference(self):
        """Tests that when a connected Stepper object is referenced to the
        negative edge of the axis through move_stage_to_ref, the final position
        is indeed the negative edge (0)."""
        self.stepper.move_stage_to_ref('FNL')
        assert isclose(self.stepper.get_curr_pos()['1'],0,rel_tol=1e-8)

    def test_positive_reference(self):
        """Tests that when a connected Stepper object is referenced to the
        positive edge of the axis through move_stage_to_ref, the final position
        is indeed the positive edge (102)."""
        self.stepper.move_stage_to_ref('FPL')
        assert isclose(self.stepper.get_curr_pos()['1'],102,rel_tol=1e-8)
        
    def test_get_velocity(self):
        """Tests that when the Stepper object is connected, the method get_velocity
        returns the correct value of velocity"""
        velocity = self.stepper.get_velocity()
        assert isclose(GCS2Commands.qVEL(self.stepper.pidevice)['1'],velocity, rel_tol = 1e-8)
    
    def test_get_acceleration(self):
        """Tests that when the Stepper object is connected, the method get_acceleration
        returns the correct value of acceleration"""
        acceleration = self.stepper.get_acceleration()
        assert isclose(GCS2Commands.qACC(self.stepper.pidevice)['1'],acceleration, rel_tol = 1e-8)
    
    def test_velocity_is_set(self):
        """Tests that when the velocity of a connected Stepper object is set through set_velocity, 
        the velocity stored in the controller's ROM is indeed the set one."""
        self.stepper.set_velocity(15.44)
        assert isclose(self.stepper.get_velocity(),15.44,rel_tol=1e-8)
    
    def test_acceleration_is_set(self):
        """Tests that when the acceleration of a connected Stepper object is set through set_acceleration, 
        the acceleration stored in the controller's ROM is indeed the set one."""
        self.stepper.set_acceleration(12.69)
        assert isclose(self.stepper.get_acceleration(),12.69,rel_tol=1e-8)
       
    def test_target_is_reached(self):
        """Tests that when the connected Stepper object is moved to a certain target through 
        move_stage_to_target, the position when the motion is completed is equal to the target."""
        self.stepper.move_stage_to_target(10)
        assert isclose(self.stepper.get_curr_pos()['1'],10,rel_tol=1e-8)
    
    def test_trigger_is_set(self):
        """Tests that when the trigger of the connected Stepper object is set through 
        configure out_trigger, the type of trigger stored in the ROM is the selected one."""
        self.stepper.configure_out_trigger(6)
        assert self.stepper.pidevice.qCTO(1, 3)[1][3] == '6'

    def test_connection_is_closed(self):
        """Tests that the connection of the Stepper object with controller is indeed closed 
        with close_connection."""
        self.stepper.close_connection()
        assert not self.stepper.pidevice.IsConnected()


class TestInputProcessor:
    scanPars = {		
			    "type": "discrete",
			    "scan_edges": [0,1],
                "stepsize" : 0.1,
			    "velocity" : 1,
		    	"acceleration" : 1,
			    "sampling_freq" : 100
		        }

    inputprocess = InputProcessor(scanPars)

    def test_initialization(self):
        """Tests that when an instance of InputProcessor is created, the inizialitation of the 
        object's attributes through __init__ method is correct."""
        assert self.inputprocess.type == "discrete"
        assert isclose(self.inputprocess.scan_edges[0],0,rel_tol = 1e-8)
        assert isclose(self.inputprocess.scan_edges[1],1,rel_tol = 1e-8)
        assert isclose(self.inputprocess.stepsize,0.1,rel_tol= 1e-8)
        assert isclose(self.inputprocess.vel,1,rel_tol=1e-8)
        assert isclose(self.inputprocess.acc,1,rel_tol=1e-8)
        assert isclose(self.inputprocess.sampl_freq,100,rel_tol=1e-8)

    def test_delta_calculation(self):
        """Tests that when an InputProcessor object is instantiated, the attribute delta 
        is correctly evaluated through delta_calculator given the input scan_edges."""
        assert isclose(self.inputprocess.delta,1,rel_tol = 1e-8)
    
    def test_delta_traslation_invariant(self):
        """Tests that when an InputProcessor object is instantiated, the attribute delta 
        is invariant under rigid traslation of the scan_edges coordinates."""
        # assumes that scan_edges[1] + 10 < 102
        self.inputprocess.scan_edges[0] += 10
        self.inputprocess.scan_edges[1] += 10
        assert isclose(self.inputprocess.delta,self.inputprocess.delta_calculator(),rel_tol=1e-8)
        
    def test_duration_positive(self):
        """ Tests that when the InputProcessor object is instantiated, the computation 
        of the duration always results in a positive (>=0, if no motion is desired) number.       
        """
        assert self.inputprocess.duration_calculator() >= 0
       
    def test_duration_velocity_not_constant(self):
        """Tests that when the InputProcessor object is instantiated, the computation 
        of the duration of the motion through duration_calculator is corrected. 
        Here is covered the case in which a constant velocity is never reached, i.e, sqrt(acc*delta)<=vel         
        """
        assert isclose(self.inputprocess.duration_calculator(),2,rel_tol = 1e-8)
    
    def test_duration_velocity_constant(self):
        """Tests that when the InputProcessor object is instantiated, the computation 
        of the duration of the motion through duration_calculator is corrected. 
        Here is covered the case in which a constant velocity is reached, i.e, sqrt(acc*delta)>vel         
        """
        self.inputprocess.acc = 0.5
        assert isclose(self.inputprocess.duration_calculator(),2*np.sqrt(2),rel_tol=1e-8)

    def test_rows_continous_values(self):
        """Tests that when an InputProcessor objected is instantiated and the scan mode is set
        to continuous, the number of rows of the DAQ must be set to 1"""
        N_rows , _ = self.inputprocess.rows_columns_contiuous()
        assert N_rows == 1
        
    def test_rows_equal_cols(self):
        """Tests that when an InputProcessor object is intantiated, the number of columns obtained in 
        continuous mode is equal to the number of rows obtained in discrete mode."""
        _ , N_cols = self.inputprocess.rows_columns_contiuous()
        N_rows, _ = self.inputprocess.rows_columns_discrete()
        assert isclose(N_rows,N_cols,rel_tol = 1e-8)
        
    def test_cols_continuous_values(self):
        """Tests that when an InputProcessor objected in instantiated and the scan mode is set
        to continuous, the number of columns is given by the floor of the ratio of delta over stepsize."""
        _ , N_cols = self.inputprocess.rows_columns_contiuous()
        assert N_cols == 11
    
    def test_cols_discrete_values(self):
        """Tests that when an InputProcessor objected in instantiated and the scan mode is set
        to discrete, the number of columns of the DAQ must be set to the floor of the product
        of the sampling frequency with the characteristic time of 50 ms."""
        _ , N_cols = self.inputprocess.rows_columns_discrete()
        assert isclose(N_cols,5,1e-8)
        
    def test_rows_cols_types(self):
        """Tests that when an InputProcessor object is instantiated the types of the number of rows
        and the number of columns (calculated either in continuous or discrete mode) are integer."""
        N_rows_cont, N_cols_cont = self.inputprocess.rows_columns_contiuous()
        N_rows_disc, N_cols_disc = self.inputprocess.rows_columns_discrete()
        tup = (N_rows_cont,N_cols_cont,N_rows_disc,N_cols_disc)
        assert all(isinstance(i, int) for i in tup)

    def test_daq_pars_is_dict(self):
        """Tests that when an InputProcessor object is instantiated, the daq_pars object 
        that is constructed with the values of the DAQ is a dictionary."""
        daq_pars = self.inputprocess.evaluate_daq_pars()
        assert isinstance(daq_pars,dict)
    
    def test_daq_pars_continuous(self):
        """Tests that when an InputProcessor object is instantiated, the scan mode is set 
        to continuous and the daq_pars dictionary is built-up, the values associated
        with the keys are corrected. """
        self.inputprocess.type = "continuous"
        daq_pars = self.inputprocess.evaluate_daq_pars()
        N_rows, N_cols = self.inputprocess.rows_columns_contiuous()
        duration = self.inputprocess.duration_calculator()
        assert daq_pars["daq_columns"] == N_cols
        assert daq_pars["daq_rows"] == N_rows
        assert daq_pars["duration"] == duration
        assert daq_pars["mode"] == "linear"
        assert daq_pars["trigger type"] == "HW trigger"
        assert daq_pars["trigger edge"] == "positive"
        assert isclose(daq_pars["holdoff"],duration*0.95,rel_tol = 1e-8)

    def test_daq_pars_discrete(self):
        """Tests that when an InputProcessor object is instantiated, the scan mode is set 
        to discrete and the daq_pars dictionary is built-up, the values associated
        with the keys are corrected. """
        self.inputprocess.type = "discrete"
        daq_pars = self.inputprocess.evaluate_daq_pars()
        N_rows, N_cols = self.inputprocess.rows_columns_discrete()
        duration = 0.05
        assert daq_pars["daq_columns"] == N_cols
        assert daq_pars["daq_rows"] == N_rows
        assert daq_pars["duration"] == duration
        assert daq_pars["mode"] == "Exact (on-grid)"
        assert daq_pars["trigger type"] == "HW trigger"
        assert daq_pars["trigger edge"] == "negative"
        assert isclose(daq_pars["holdoff"],duration*0.95,rel_tol = 1e-8)


class TestScanners:
    inPars = {
        "scan_pars" : {		
            "type": "discrete",
            "scan_edges": [0,1],
            "stepsize" : 0.1,
            "velocity" : 1,
            "acceleration" : 1,
            "sampling_freq" : 100
            },
        "pi" : 	{	
            "ID":"C-663",
            "stage_ID": "L-406.40SD00",
            "refmode": "FNL",
            "trig_type":6
		    }
            }
    
    scanner = Scanner(inPars)
    
    def test_initialization_pars(self):
        """Tests that when an instance of Scan1D is created, the inizialitation of the 
        object's attributes through __init__ method is correct."""
        assert self.scanner.PI["ID"] == "C-663"
        assert self.scanner.PI["stage_ID"] == "L-406.40SD00"
        assert self.scanner.PI["refmode"] == "FNL"
        assert self.scanner.PI["trig_type"] == 6
        assert self.scanner.scan_pars["type"] == "discrete"
        assert isclose(self.scanner.scan_pars["scan_edges"][0],0,rel_tol = 1e-8)
        assert isclose(self.scanner.scan_pars["scan_edges"][1],1,rel_tol = 1e-8)
        assert isclose(self.scanner.scan_pars["stepsize"],0.1,rel_tol = 1e-8)
        assert isclose(self.scanner.scan_pars["velocity"],1,rel_tol= 1e-8)
        assert isclose(self.scanner.scan_pars["acceleration"],1,rel_tol= 1e-8)
        assert isclose(self.scanner.scan_pars["sampling_freq"],100,rel_tol= 1e-8)
    
    def test_dimension_target_positions(self):
        """Tests that when an instance of Scan1D is created,the dimension of the 
        targets array is correct." 
        """
        N_targets = 11
        assert len(self.scanner.targets) == N_targets
        
    def test_values_target_positions(self):
        """Tests that when an instance of Scan1D is created, the first and the final 
        point of targets is correct."""
        assert isclose(self.scanner.targets[0],self.scanner.scan_edges[0],rel_tol = 1e-8)
        assert isclose(self.scanner.targets[-1],self.scanner.scan_edges[1],rel_tol = 1e-8)  # by induction is right because of linspace
    
    def test_stepper_is_instantiated(self):
        """Tests that when an instance of Scan1D is created, the instance of Stepper
        is properly instantiated throught the __init__ method."""
        assert isinstance(self.scanner.stepper,Stepper)
    
    def test_stepper_is_connected(self):
        """Tests that when an instance of Scan1D is created and connected_master is called, 
        master is properly connected to the PI controller"""
        self.scanner.connect_stepper()
        assert self.scanner.stepper.pidevice.isConnected()

    def test_vel_acc_stepper(self):
        """Tests that when an instance of Scan1D is created and stepper is connected, 
        velocity and acceleration are properly set through the method setup_motion_stepper"""
        self.scanner.setup_motion_stepper(self)
        assert isclose(self.scanner.stepper.get_velocity(),1,rel_tol = 1e-8)
        assert isclose(self.scanner.stepper.get_acceleration(),1,rel_tol = 1e-8)
        
    def test_initial_scan_position(self):
        """Tests that when an instance of Scan1D is created and init_stepper_scan is called,
        the axis is moved to the first target position."""
        self.scanner.init_stepper_scan()
        cur_pos = self.scanner.stepper.get_curr_pos()
        assert isclose(cur_pos,self.scanner.targets[0])

    def test_trigger_is_configured(self):
        """Tests that when an instance of Scan1D is created and init_1D_scan is called, 
        the output trigger is set and activated.
        """
        trig_type = self.inPars ["pi"]["trig_type"]
        assert self.scanner.stepper.pidevice.qCTO(1, 3)[1][3] == str(trig_type)
        
    def test_discrete_scan(self):
        """Tests that when an instance of Scan1D is created and execute_discrete_scan
           is performed, the covered positions are the targeted ones."""
        self.scanner.init_stepper_scan()
        cur_pos = self.scanner.execute_discrete_scan()
        assert all(abs(pos - target) <= 0.5e-3 for pos, target in zip(cur_pos, self.scanner.targets))
    
    def test_continuous_1D_scan(self):
        """Tests that when an instance of Scan1D is created and execute_continuous_scan
        is performed, the position of the stage is equal to the last target point.
        """
        self.scanner.init_stepper_scan()
        self.scanner.execute_continuous_scan()
        cur_pos = self.scanner.stepper.get_curr_pos()
        assert isclose(cur_pos,self.scanner.targets[-1])
        
class TestOutputProcessor:
    scan_pars = {		
            "type": "discrete",
            "scan_edges": [0.3,0.],
            "stepsize" : 0.001,
            "velocity" : 1,
            "acceleration" : 1,
            "sampling_freq" : 100
            }
    daq_pars = {
                "daq_columns" : 100,
                "daq_rows" : 300,
                "duration" : 0.05,
                "mode" : "Exact (on-grid)",
                "trigger type" : "HW trigger",
                "trigger edge" : "negative",
                "holdoff" : 0.05*(0.95),
            }     
    
    filename = "dev4910_demods_0_sample_r_avg_00000.csv"
    outputprocess = OutputProcessor(filename,scan_pars,daq_pars)
    
    def test_initialization_scan_pars(self):
        """Tests that when an OutputProcessor object is instantiated, the attributes inizialization
        through the __init__ method is correct"""
        assert self.scan_pars == self.outputprocess.scan_pars
        assert self.filename == self.outputprocess.filename
        assert self.daq_pars["daq_columns"] == self.outputprocess.N_cols
        assert self.daq_pars["daq_rows"] == self.outputprocess.N_rows
        
    def test_raw_data_shape(self):
        """Tests that when an OutProcessor object is instantiated and get_raw_data is called,
        the returned NumPy has the correct number of rows"""
        raw_data = np.genfromtxt(self.filename,skip_header = 1, delimiter = ";")
        returned_column = self.outputprocess.get_raw_data()
        assert raw_data.shape[0] == returned_column.shape[0]

    def test_raw_data_values(self):
        """Tests that when an OutProcessor object is instantiated and get_raw_data is called,
        the returned NumPy has the same values of the original one."""
        # find raw data column directly, select only non nan values
        raw_data = np.genfromtxt(self.filename,skip_header = 1, delimiter = ";")
        raw_column = raw_data[:,2]
        raw_column_not_nan = raw_column[~np.isnan(raw_column)]
        # find raw data through get_raw_data, select only non nan values
        returned_column = self.outputprocess.get_raw_data()
        returned_column_not_nan = returned_column[~np.isnan(returned_column)]
        assert all(abs(raw - returned)/raw <= 1e-8 for raw, returned in zip(raw_column_not_nan,returned_column_not_nan))
    
    def test_dimension_averaged_subintervals_is_rows(self):
        """Tests tat when an OutputProcessor object is instantiated and evaluate_averaged_data is called, 
        the dimension of the returned array (with averaged subintervals) is equal to the 
        number of rows"""
        returned_column = self.outputprocess.get_raw_data()
        avg_returned_column = self.outputprocess.evaluate_averaged_data(returned_column)
        N_rows_avg = len(avg_returned_column)
        assert self.daq_pars["daq_rows"] == N_rows_avg
        
    
    def test_average_data(self):
        """Tests that when an OutputProcessor object is instantiated and evaluate_average_data
        is called, the average of the averaged subintervals (of N_cols dimensions) is equal to the
        average of the overall raw_data. This is justified by subintervals of equal dimension"""
        # find raw data column directly, select only non nan values
        raw_data = np.genfromtxt(self.filename,skip_header = 1, delimiter = ";")
        raw_column = raw_data[:,2]
        raw_column_not_nan = raw_column[~np.isnan(raw_column)]
        # find raw data through get_raw_data, select only non nan values
        returned_column = self.outputprocess.get_raw_data()
        returned_column_not_nan = returned_column[~np.isnan(returned_column)]        
        # average subintervals of returned_column_not_nan
        avg_returned_column = self.outputprocess.evaluate_averaged_data(returned_column_not_nan)
        avg_raw= np.mean(raw_column_not_nan)

        # average averaged subintervals of returned_column_not_nan
        avg_avg_returned = np.mean(avg_returned_column)
          
        assert isclose(avg_avg_returned,avg_raw,rel_tol = 1e-12)
        
    def test_dimension_target_positions(self):
        """Tests that when an instance of OutputProcessor is created and evaluate_target_position is called,
        the dimension of the targets array is correct."""
        scanedges = self.scan_pars["scan_edges"]
        stepsize = self.scan_pars["stepsize"]
        targets = self.outputprocess.evaluate_target_positions(scanedges,stepsize)
        N_targets = 301
        assert N_targets == len(targets)
        
    def test_values_target_positions(self):
        """Tests that when an instance of OutputProcessor is created and evasluate_target_position is called,
        the first and the last point of the target array is corrected"""
        scanedges = self.scan_pars["scan_edges"]
        stepsize = self.scan_pars["stepsize"]
        targets = self.outputprocess.evaluate_target_positions(scanedges,stepsize)

        assert isclose(scanedges[0],targets[0],rel_tol = 1e-8)
        assert isclose(scanedges[1],targets[-1],rel_tol = 1e-8)  # by induction is right because of linspace

    
    def test_save_data_file(self):
        """Tests that when an instance of OutputProcessor is created and save_data_file  is called, 
        an output file named "cleaned_1D_data.txt" is indeed saved in the current folder"""
        # create fictitious targets 
        targets  = np.array([1,2,3,4,5],dtype=float)
        # create fictitious signal values
        avg_val = np.random.random(size = len(targets))
        # save file to "cleaned_1D_data.txt"
        self.outputprocess.save_data_file(targets,avg_val)
        # expected output
        expected_output = ''
        for i in range(len(targets)):
            expected_output += '\b'+str(targets[i])+', '+'{:.6f}'.format(avg_val[i])+'\n'

        # Read the file contents and compare with expected output
        with open("cleaned_1D_data.txt", "rb") as f:
            assert f.read(),expected_output
            
        
class TestInputValidator:
    """ A class to test that when invalid input scan parameters are inserted, an exeption is raised and the code stops the execution. 
     Although a lower-level control is already present in the internal PI subroutines, it takes time to be activated. 
     Therefore, this additional entrance control find out invalid inputs without the need of connecting to the PI instrument. 
    """

    scanPars = {		
        "type": "discrete",
        "scan_edges": [0,1],
        "stepsize" : 0.1,
        "velocity" : 1,
        "acceleration" : 1,
        "sampling_freq" : 100
        }
    
    
    inputval = InputValidator(scan_pars=scanPars)
      
    def test_valid_input(self):
        """Tests that when an InputValidator class is instantiated and corrected values of scan parameters
        are provided, the validate method does not raise any error and return True."""
        assert self.inputval.validate() == True

    def test_invalid_type(self):
        """Tests that when an InputValidator class is instantiated and an invalid value of 'type' is passed 
        to the validate_type function, an exception is raised."""
        self.inputval.type = "invalid"
        with pytest.raises(ValueError):
            self.inputval.validate_type()

    def test_invalid_scan_edges(self):
        """ Tests that when an InputValidator class is instantiated an invalid value of 'scan_edges' is passed 
        to the validate_scan_edges function, an exception is raised """
        self.inputval.scan_edges = [-10, 110]
        with pytest.raises(ValueError):
            self.inputval.validate_scan_edges()
    
    def test_invalid_stepsize(self):
        """ Tests that when an InputValidator class is instantiated an invalid value of 'stepsize' is passed 
        to the validate_stepsize function, an exception is raised """
        self.inputval.stepsize = 200
        with pytest.raises(ValueError):
            self.inputval.validate_stepsize()

    def test_invalid_velocity(self):
        """ Tests that when an InputValidator class is instantiated an invalid value of 'velocity' is passed 
        to the validate_velocity function, an exception is raised """
        self.inputval.velocity = -5
        with pytest.raises(ValueError):
            self.inputval.validate_velocity()

    def test_invalid_acceleration(self):
        """ Tests that when an InputValidator class is instantiated an invalid value of 'acceleration' is passed 
        to the validate_velocity function, an exception is raised """
        self.inputval.acceleration = 30
        with pytest.raises(ValueError):
            self.inputval.validate_acceleration()



import numpy as np 

class OutputProcessor():

    def __init__(self,filename,scan_pars,daq_pars):
        """
        Initialize an OutputProcessor object with the given filename, scan parameters, and DAQ parameters.

        Args:
        - filename (str): the name of the file containing the output data
        - scan_pars (dict): a dictionary containing the scan parameters
        - daq_pars (dict): a dictionary containing the DAQ parameters

        Attributes:
        - scan_pars (dict): a dictionary containing the scan parameters
        - filename (str): the name of the file containing the output data
        - daq_pars (dict): a dictionary containing the DAQ parameters
        - N_rows (int): the number of rows in the DAQ data
        - N_cols (int): the number of columns in the DAQ data
        """
        self.scan_pars = scan_pars
        self.filename = "output/"+filename
        self.N_rows = daq_pars["daq_rows"]
        self.N_cols = daq_pars["daq_columns"]

    def get_raw_data(self):
        """
        Read raw data from the outputfile and return the third column, containing the values of the measured signal.

        Returns:
        - raw_data (ndarray): a NumPy array containing the third column of the input file
        """
        raw_data = np.genfromtxt(self.filename,skip_header = 1, delimiter = ";")
        return raw_data[:,2] # third column contains all the data in a column     
    
    def evaluate_averaged_data(self,raw_data): 
        """
        Average the input data in case of a 1D discrete scan. Raw data is divided ini chunck of dimension N_cols, 
        over which an average is performed. Output avg_data is in one to one relation with the spatial coordinates.

        Args:
        - raw_data (ndarray): a NumPy array containing the raw data

        Returns:
        - avg_data (ndarray): a NumPy array containing the averaged data
        """
        avg_data = np.empty(self.N_rows)
        for row in range(self.N_rows):
            avg_data[row] = np.mean(raw_data[row*self.N_cols:(row + 1)*self.N_cols])
        return avg_data
    
    def evaluate_target_positions(self,scanedges,stepsize):
        """
        Evaluate the partition of the target points for a 1D scan.

        Args:
        - scanedges (list): a list containing the start and end positions of the scan
        - stepsize (float): the step size of the scan

        Returns:
        - targets (ndarray): a NumPy array containing the target positions
        """
        # calculate targets points
        Npoints = int(abs((scanedges[1]-scanedges[0]))/stepsize) + 1
        return np.linspace(scanedges[0],scanedges[1],Npoints,endpoint=  True)

    def save_data_file(self,targets,avg_data):         
        """
        Save the cleaned 1D data to a file named "cleaned_1D_data.txt".

        Args:
        - targets (ndarray): a NumPy array containing the target positions
        - avg_data (ndarray): a NumPy array containing the averaged data
        """
        out_name = "output/cleaned_1D_data.txt"
        out_file = np.column_stack((targets,avg_data))
        np.savetxt(out_name, out_file, fmt = "%10.6f", delimiter = ",")

    def save_processed_data(self):
        """
        Process the 1D data, averaging it if necessary (discrete scan), and save it to a file named "cleaned_1D_data.txt".
        """
        # get targets position
        scanedges = self.scan_pars["scan_edges"]
        stepsize = self.scan_pars["stepsize"]
        targets = self.evaluate_target_positions(scanedges,stepsize)  
        # get out_data
        raw_data = self.get_raw_data()
        if (self.scan_pars["type"] == "discrete"):
            out_data = self.evaluate_averaged_data(raw_data, )
        else: 
            out_data = raw_data
        # save data
        self.save_data_file(targets,out_data)

            
        
               
        
        


    




from pizurscan.PI_commands import Stepper
import numpy as np

class Scanner:
    """
    A class for performing a 1D scan using a PI device.

    Attributes:
        PI (dict): Dictionary of PI device information.
        scan_pars (dict): Dictionary of scan parameters.
        scan_edges (tuple): Tuple of start and end positions of the scan.
        stepsize (float): Step size of the scan.
        targets (numpy.ndarray): Array of target positions for the scan.
        stepper (Stepper): Stepper object for controlling the PI device.

    Methods:
        __init__(self, InPars):
            Initializes Scan1D object with input parameters.
        evaluate_target_positions(self):
            Evaluates the partition of the target points for a 1D scan.
            Returns:
                numpy.ndarray: Array of target positions.
        connect_stepper(self):
            Connects to the PI device through a user-interface I/O.
        setup_motion_stepper(self):
            Stores input velocity and acceleration in the ROM of the device.
        init_stepper_scan(self):
            Sets up the 1D scan in four steps.
        execute_discrete_scan(self):
            Executes the 1D discrete scan by moving the axis on all the target positions.
            Returns:
                list: List of current positions.
        execute_continuous_scan(self):
            Executes the continuous scan by moving the axis to the last position.
    """

    def __init__(self, InPars):
        """ Initializes Scanner object with input parameters.
        
        Parameters:
        ----------
        - InPars : dict
            a dictionary of input parameters regarding the scan features
        
        Attributes:
        ----------
        - PI : dict
            a dictionary containing the PI controller and axis id
        - scan_pars: dict
            a dictionary containing the scan parameters
        - scan_edges : list
            a list containing the two edges of the scan. Axis will move from the leftward to the rightward.
        - stepsize : float
            a float containing the step size of the scan
        - targets : numpy.array
            a numpy.array containing the targets positions of the scan
        - stepper: Stepper
            a stepper object that instantiate Stepper class.
        """
        self.PI = InPars["pi"]
        self.scan_pars = InPars["scan_pars"]
        self.scan_edges = self.scan_pars["scan_edges"]
        self.stepsize = self.scan_pars["stepsize"]
        self.targets = self.evaluate_target_positions()
        self.stepper = Stepper(self.PI["ID"], self.PI["stage_ID"])

    def evaluate_target_positions(self):
        """
        Evaluates the partition of the target points for a 1D scan.

        Returns:
        --------
        numpy.ndarray: Array of target positions.
        """
        Npoints = int(abs(self.scan_edges[1] - self.scan_edges[0]) / self.stepsize) + 1
        return np.linspace(self.scan_edges[0], self.scan_edges[1], Npoints, endpoint=True)

    def connect_stepper(self):
        """
        Connects to the PI device through a user-interface I/O.
        """
        self.stepper.connect_pidevice()

    def setup_motion_stepper(self):
        """
        Stores input velocity and acceleration in the ROM of the device.
        """
        self.stepper.set_velocity(self.scan_pars["velocity"])
        self.stepper.set_acceleration(self.scan_pars["acceleration"])
        
    def init_stepper_scan(self):
        """Setup the 1D scan in four steps: (1) acceleration and velocity are set to
        default values for quick refering and motion. (2) Stage is moved to reference,
        either positive or negative depending on the input refmode. (3) stage is moved 
        to the first target point output trigger is selected and activated
        """
        # high default values to obtain a quick referencing 
        self.stepper.set_velocity(10)
        self.stepper.set_acceleration(20)
        self.stepper.move_stage_to_ref(self.PI["refmode"])
        self.stepper.move_stage_to_target(self.targets[0])
        trigtype = self.PI["trig_type"]
        self.stepper.configure_out_trigger(trigger_type=trigtype)

    def execute_discrete_scan(self):
        """
        Execute the 1D discrete scan by moving the axis on all the target positions.

        Returns:
        --------
        list : List of current positions.
        """
        cur_pos = []
        for target in self.targets:
            self.stepper.move_stage_to_target(target)        
            print("Position: ", target)
            cur_pos.append(self.stepper.get_curr_pos())
        self.stepper.close_connection()
        return cur_pos

    
    def execute_continuous_scan(self):
        """
        Execute the continuous scan by moving the axis to the last position.
        """
        self.stepper.move_stage_to_target(self.targets[-1])    
        self.stepper.close_connection()
    
        